/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>

/*
this is the main function if the listener is to manually connect to etherx via a socket
*/

irctrans_data ircdata;

result irc_phandler_receive(instance id, dpacket dp, void *arg)
{
    irctrans_data ircdata = (irctrans_data)arg;
    jpacket jp;

    log_debug(ZONE, "EXTRA_DEBUG: irc_phandler_recieve: %s", xmlnode2str(dp->x));
    /* strip the route off */
    if(dp->type == p_ROUTE)
    {
        jutil_error(dp->x, TERROR_NOTALLOWED); /* we do not accept routes */
        deliver(dpacket_new(dp->x), NULL);
        return r_DONE;
    }

    jp = jpacket_new(dp->x);

    if(jp == NULL || jp->type == JPACKET_UNKNOWN)
    {
        xmlnode_free(jp->x);
        return r_DONE;
    }

    jp->aux1 = (void*)ircdata;
    log_debug(ZONE, "EXTRA_DEBUG: handing of to mtq");
    mtq_send(ircdata->q, jp->p, irc_parser, (void*)jp);
    return r_DONE;
}

void _irc_trans_stop(xht sessions, const char *key, void *val, void *arg)
{
    xmlnode m = xmlnode_new_tag("message");
    irc_session s = (irc_session)val;
    irctrans_data ircdata = (irctrans_data)arg;

    xmlnode_put_attrib(m, "type", "error");
    xmlnode_put_attrib(m, "from", jid_full(s->j));
    xmlnode_put_attrib(m, "to", ircdata->i->id);

    jutil_error(m, TERROR_EXTERNAL);
    xmlnode_insert_cdata(xmlnode_insert_tag(m, "body"), "IRC Transport is shutting down", -1);

    deliver(dpacket_new(m), NULL);

    mio_close(s->m);
}

/* used to exit the server and close down all session gracefully */
void irc_trans_stop(void *arg)
{
    irctrans_data ircdata = (irctrans_data)arg;

    xhash_walk(ircdata->sessions, _irc_trans_stop, (void*)ircdata);
    xhash_free(ircdata->sessions);
}
/*
int handle_config(xmlnode conf)
{
    xmlnode cur;

    for(cur = xmlnode_get_firstchild(conf); cur != NULL; cur = xmlnode_get_nextsibling(cur))
    {
        if(xmlnode_get_type(cur) != NTYPE_TAG)
            continue;

        if(j_strcmp(xmlnode_get_name(cur), "register") == 0)
            ircdata->allow_register = 1;
        else if(j_strcmp(xmlnode_get_name(cur), "default") == 0)
        {
            ircdata->default_server = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "default/server"));
            ircdata->default_port   = pstrdup(xmlnode_pool(conf), xmlnode_get_tag_data(conf, "default/port"));
        }
        else if(j_strcmp(xmlnode_get_name(cur),"quit")==0)
        { 
            users first, curr;
            xmlnode list;
            first = quit_list;   
            list = xmlnode_get_firstchild(cur);
            while(list != NULL && j_strcmp(xmlnode_get_name(list), "message") != 0)
                list = xmlnode_get_nextsibling(list);    
            if(list != NULL)
            {
                quit_list = pmalloc(main_pool, sizeof(_users));
                first = quit_list;
                first->user = pstrdup(main_pool, xmlnode_get_data(xmlnode_get_firstchild(list)));
                first->next = first; 
                while((list = xmlnode_get_nextsibling(list)) != NULL)
                {
                    if(j_strcmp(xmlnode_get_name(list), "message") == 0)    
                    {
                        curr = pmalloc(main_pool, sizeof(_users));   
                        curr->user = pstrdup(main_pool, xmlnode_get_data(xmlnode_get_firstchild(list)));
                        quit_list->next = curr;
                        quit_list = quit_list->next;
                        quit_list->next = first;
                    }
                }
            }
        }
        else if(j_strcmp(xmlnode_get_name(cur),"part")==0)
        {
            users first,curr;
            xmlnode list;
            first=quit_list;   
            list=xmlnode_get_firstchild(cur);
            while(list!=NULL&&j_strcmp(xmlnode_get_name(list),"message")!=0)
                list=xmlnode_get_nextsibling(list);    
            if(list!=NULL)
            {
                part_list=pmalloc(main_pool,sizeof(_users));
                first=part_list;
                first->user=pstrdup(main_pool,xmlnode_get_data(xmlnode_get_firstchild(list)));
                first->next=first; 
                while((list=xmlnode_get_nextsibling(list))!=NULL)
                {
                    if(j_strcmp(xmlnode_get_name(list),"message")==0)    
                    {
                        curr=pmalloc(main_pool,sizeof(_users));   
                        curr->user=pstrdup(main_pool,xmlnode_get_data(xmlnode_get_firstchild(list)));
                        part_list->next=curr;
                        part_list=part_list->next;
                        part_list->next=first;
                    }
                }
            }
        }
    }

    if(ircdata->default_port == NULL)
        ircdata->default_port = pstrdup(main_pool, "6667");
    if(ircdata->default_server == NULL)
        ircdata->default_server = pstrdup(main_pool, "irc.linux.com");
    return 0;
}
*/
/* load up the irc transport from jabberd */
void irc_transport(instance i, xmlnode x)
{
    xmlnode config;

    log_notice(i->id, "IRC Transport starting up...");
    log_debug(ZONE, "EXTRA_DEBUG: starting up");

    ircdata            = pmalloco(i->p, sizeof(_irctrans_data));
    ircdata->i         = i;
    ircdata->xc        = xdb_cache(i);
    ircdata->q         = mtq_new(i->p);
    ircdata->cfg       = xdb_get(ircdata->xc, jid_new(xmlnode_pool(x), "config@-internal"), "jabberd:irc:config");
    ircdata->sessions  = xhash_new(101);
    ircdata->browse    = xmlnode_new_tag("query");
    ircdata->start     = time(NULL);

    xmlnode_put_attrib(ircdata->browse, "xmlns", "jabber:iq:browse");
    xmlnode_put_attrib(xmlnode_insert_tag(ircdata->browse, "service"), "type", "irc");
    xmlnode_put_attrib(xmlnode_get_tag(ircdata->browse, "service"), "jid", i->id);
    xmlnode_put_attrib(xmlnode_get_tag(ircdata->browse, "service"), "name", xmlnode_get_tag_data(ircdata->cfg, "vCard/FN"));
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_REGISTER, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_VERSION, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_BROWSE, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_TIME, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_LAST, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_VCARD, -1);
    xmlnode_insert_cdata(xmlnode_insert_tag(xmlnode_get_tag(ircdata->browse, "service"), "ns"), NS_CONFERENCE, -1);

    log_debug(ZONE, "EXTRA_DEBUG: created browse packet: %s", xmlnode2str(ircdata->browse));
    log_debug(ZONE, "IRC Transport Loading Configuration: %s", xmlnode2str(config));

    register_phandler(i, o_DELIVER, irc_phandler_receive, (void*)ircdata);

    if(xmlnode_get_tag(ircdata->cfg, "update") != NULL)
    {
        xmlnode x;
        spool sp;
        log_debug(ZONE,"sending out update query");
        x = xmlnode_new_tag("presence");
        sp = spool_new(xmlnode_pool(x));
        xmlnode_put_attrib(x, "to", spools(xmlnode_pool(x),"964735456@", xmlnode_get_tag_data(ircdata->cfg, "update/host"), "/", IRC_T_VERSION, xmlnode_pool(x)));
        xmlnode_put_attrib(x, "from", ircdata->i->id);
        deliver(dpacket_new(x), i);
    }

    pool_cleanup(i->p, irc_trans_stop, (void*)ircdata);
}
