/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>

void irc_add_to_buffer(irc_session s, char *buffer, int bufsz)
{
    char *temp;

    if(s->buffer == NULL)
    {
        s->buffer = malloc(bufsz + 1);
        s->buffer[0] = '\0';
        strcat(s->buffer, buffer);
        return;
    }

    temp = s->buffer;
    s->buffer = malloc(strlen(temp) + bufsz +1);

    s->buffer[0] = '\0';
    strcat(s->buffer, temp);
    strcat(s->buffer, buffer);

    free(temp);
}

char *irc_get_next_line(irc_session s, pool p)
{
    char *ptr;
    char *ret;

    if(s->buffer == NULL)
        return NULL;

    ptr = strchr(s->buffer, '\r');

    if(ptr == NULL)
        ptr = strchr(s->buffer, '\n');

    if(ptr == NULL)
        return NULL;

    ptr[0] = '\0';
    ptr++;

    while(ptr[0] == '\r' || ptr[0] == '\n') ptr++;

    ret = pstrdup(p, s->buffer);

    if(ptr[0] != '\0')
    {
        char *temp = s->buffer;
        s->buffer = malloc(strlen(ptr) + 1);
        s->buffer[0] = '\0';
        strcat(s->buffer, ptr);
        free(temp); 
    }
    else
    {
        free(s->buffer);
        s->buffer = NULL;
    }

    return ret;
}

void __irc_cleanup_channel(xht users, const char *key, void *val, void *arg)
{
    user u = (user)val;
    log_debug(ZONE, "ZAPPING user %s key %s id %s", u->user, key, jid_full(u->id));

    if(u->ops == -1) u->ops = -2;
    else u->ops = -1;

    xhash_zap(users, u->user);
    xhash_zap(users, jid_full(u->id));

    if(u->ops == -2)
        pool_free(u->p);
}

void _irc_cleanup_channel(xht channels, const char *key, void *val, void *arg)
{
    chans channel = (chans)val;
    irc_session s = (irc_session)channel->data;
    irctrans_data ircdata = (irctrans_data)s->data;

    log_notice(ircdata->i->id, "Shutting Down Channel %s on server %s for user %s", channel->name, s->server, jid_full(s->j));

    xhash_walk(channel->users, __irc_cleanup_channel, arg);
    xhash_free(channel->users);

    if(channel->data == (void*)-1) channel->data = (void*)-2;
    else channel->data = (void*)-1;

    xhash_zap(channels, channel->id);
    xhash_zap(channels, channel->name);

    if(channel->data == (void*)-2)
        pool_free(channel->p);
}

void __irc_check_quit(xht users, const char *key, void *val, void *arg)
{
    irc_session s;
    chans channel;

    if(j_strcmp(key, (char*)arg) == 0)
    {
        user u = (user)val;
        jpacket jp = jpacket_new(xmlnode_new_tag("iq"));
        xmlnode x;

        channel = (chans)u->data;
        s = (irc_session)channel->data;

        xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
        xmlnode_put_attrib(jp->x, "from", channel->id);
        xmlnode_put_attrib(jp->x, "type", "set");
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_BROWSE);
        jpacket_reset(jp);

        x = xmlnode_insert_tag(jp->iq, "user");
        xmlnode_put_attrib(x, "jid", jid_full(u->id));
        xmlnode_put_attrib(x, "type", "remove");

        xhash_zap(users, key);
        xhash_zap(users, jid_full(u->id));
        pool_free(u->p);

        deliver(dpacket_new(jp->x), NULL);
    }
}

void _irc_check_quit(xht channels, const char *key, void *val, void *arg)
{
    chans channel = (chans)val;
    xhash_walk(channel->users, __irc_check_quit, arg);
}

typedef struct irc_nick_change_st
{
    char *old_nick;
    char *new_nick;
} _irc_nick_change, *irc_nick_change;

void __update_nick(xht users, const char *key, void *val, void *arg)
{
    irc_session s;
    chans channel;
    irc_nick_change nc = (irc_nick_change)arg;

    if(j_strcmp(key, nc->old_nick) == 0)
    {
        user u = (user)val;
        jpacket jp = jpacket_new(xmlnode_new_tag("iq"));
        xmlnode x;

        channel = (chans)u->data;
        s = (irc_session)channel->data;

        xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
        xmlnode_put_attrib(jp->x, "from", channel->id);
        xmlnode_put_attrib(jp->x, "type", "set");
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_BROWSE);
        jpacket_reset(jp);

        x = xmlnode_insert_tag(jp->iq, "user");
        xmlnode_put_attrib(x, "jid", jid_full(u->id));
        xmlnode_put_attrib(x, "name", u->user);

        xhash_zap(users, key);

        u->user = make_lowercase(pstrdup(u->p, nc->new_nick));
        xhash_put(users, u->user, u);

        deliver(dpacket_new(jp->x), NULL);
    }
}

void _update_nick(xht channels, const char *key, void *val, void *arg)
{
    chans channel = (chans)val;
    xhash_walk(channel->users, __update_nick, arg);
}

/* normal group communications */
void irc_parse_phase_two(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    irc_session s = (irc_session)arg;
    irctrans_data ircdata = (irctrans_data)s->data;
    char *line;
    pool p;

    log_debug(ZONE, "in phase two: m%X flag:%d, arg:%X, buffer: %s, bufsz:%d", m, flag, arg, buffer, bufsz);
    if(flag == MIO_ERROR)
    {
        xmlnode x;
        while((x = mio_cleanup(m)) != NULL)
            deliver_fail(dpacket_new(x), "IRC Connection Failed");
        return;
    }

    if(flag == MIO_CLOSED)
    {
        xhash_zap(ircdata->sessions, spools(s->p, jid_full(s->j), "/", s->server, s->p));
        xhash_walk(s->channels, _irc_cleanup_channel, NULL);
        xhash_free(s->channels);
        if(s->buffer != NULL)
            free(s->buffer);
        pool_free(s->p);
        return;
    }

    irc_add_to_buffer(s, buffer, bufsz);

    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    {
        char *from = NULL, *cmd = NULL, *next = NULL;

        if(*line == '\0')
            continue;

        if(*line != ':')
        {
            cmd = line;
        }
        else
        {  
            from = line;
            cmd = strchr(line, ' ');
            if(cmd == NULL)
                continue;
            cmd[0] = '\0';
            cmd++;
        }

        next = strchr(cmd, ' ');
        if(next == NULL)
            continue;

        next[0] = '\0';
        next++;

        if(j_strcmp(cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", next, "\n", p), -1);
        }
        if(j_strcmp(cmd, "JOIN") == 0)
        {
            user u;
            pool p2;
            chans channel;
            jpacket jp;
            xmlnode x;
            char *ptr;

            ++next;
            channel = xhash_get(s->channels, make_lowercase(next));
            if(from == NULL || channel == NULL)
                continue;

            ptr = strchr(++from, '!');
            ptr[0] = '\0';
            p2 = pool_new();
            u = pmalloco(p2, sizeof(_user));
            u->p = p2;
            u->user = make_lowercase(pstrdup(u->p, from));
            u->id   = jid_new(u->p, spools(p, channel->id, "/", shahash(u->user), p));
            u->data = (void*)channel;
            xhash_put(channel->users, jid_full(u->id), u);
            xhash_put(channel->users, u->user, u);

            jp = jpacket_new(xmlnode_new_tag("iq"));
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_BROWSE);
            xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
            xmlnode_put_attrib(jp->x, "from", channel->id);
            xmlnode_put_attrib(jp->x, "type", "set");

            jpacket_reset(jp);
            x = xmlnode_insert_tag(jp->iq, "user");
            xmlnode_put_attrib(x, "jid", jid_full(u->id));
            xmlnode_put_attrib(x, "name", u->user);
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(j_strcmp(cmd, "NICK") == 0)
        {
            char *ptr;
            irc_nick_change nc;

            ptr = strchr(++from, '!');
            if(ptr == NULL) continue;

            ptr[0] = '\0';

            if(from == s->nick)
            {
                /* this is our nick change, we must have a jpacket on s->aux1 */
                jpacket jp = s->aux1;
                s->aux1 = NULL;
                if(jp == NULL)
                    continue; /* something has gone wrong */
                jutil_iqresult(jp->x);
                deliver(dpacket_new(jp->x), NULL);
            }

            nc = pmalloco(p, sizeof(_irc_nick_change));
            nc->old_nick = make_lowercase(from);
            nc->new_nick = make_lowercase(++next);
            xhash_walk(s->channels, _update_nick, (void*)nc);
        }
        else if(j_strcmp(cmd, "433") == 0) /* NICK IN USE */
        {
            jpacket jp = s->aux1;
            s->aux1 = NULL;

            if(jp == NULL)
                continue;

            if(xmlnode_get_tag_data(jp->iq, "nick") == NULL)
            {
                /* no more nicks to try */
                jutil_error(jp->x, TERROR_CONFLICT);
                deliver(dpacket_new(jp->x), NULL);
                return;
            }

            mio_write(m, NULL, spools(p, "NICK :", xmlnode_get_tag_data(jp->iq, "nick"), "\n", p), -1);
            xmlnode_hide(xmlnode_get_tag(jp->iq, "nick"));
        }
        else if(j_strcmp(cmd, "MODE") == 0)
        {
        }
        else if(j_strcmp(cmd, "PRIVMSG") == 0)
        {
            /* message from the wonderfull world of IRC */
            chans channel;
            xmlnode x;
            char *ptr,*msg;
            user u;

            ptr = strchr(next, ' ');
            if(ptr == NULL) continue;
            ptr[0] = '\0';
            msg = ptr+1;

            ptr = strchr(++from, '!');
            if(ptr == NULL) continue;
            ptr[0] = '\0';

            if(j_strcasecmp(next, s->nick) == 0)
            {
                mio_write(m, NULL, spools(p, "PRIVMSG ", from, " :Sorry, I do not support Private Messages right now\n", p), -1);
                continue;
            }

            channel = xhash_get(s->channels, make_lowercase(next));
            if(channel == NULL) 
            {
                    log_debug(ZONE, "CHANNEL NOT FOUND IN HASH FOR PRIVMSG: %s", next);
                    continue;
            }


            u = xhash_get(channel->users, make_lowercase(from)); 
            if(u == NULL) 
            {
                    log_debug(ZONE, "USER NOT FOUND IN HASH: %s", from);
                    continue;
            }

            x = xmlnode_new_tag("message");
            xmlnode_put_attrib(x, "to", jid_full(s->j));
            xmlnode_put_attrib(x, "from", jid_full(u->id));
            xmlnode_put_attrib(x, "type", "groupchat");

            next = ++msg;
            if(*next == '\001')
            {
                char *ptr;
                next++;
                if(j_strncasecmp(next, "ACTION", 6) == 0)
                {
                    ptr = strchr(next, '\001');
                    ptr[0] = '\0';
                    next +=3;
                    next[0] = '/';
                    next[1] = 'm';
                    next[2] = 'e';
                }
                else
                {
                    xmlnode_free(x);
                    continue;
                }
            }
            irc_strip_mirc_codes(next);
            next = str_to_UTF8(xmlnode_pool(x), next);
            xmlnode_insert_cdata(xmlnode_insert_tag(x, "body"), next, -1);

            deliver(dpacket_new(x), NULL);
        }
        else if(j_strcmp(cmd, "QUIT") == 0)
        {
            char *ptr;

            ptr = strchr(++from, '!');
            if(ptr == NULL) continue;

            ptr[0] = '\0';

            from = make_lowercase(from);

            xhash_walk(s->channels, _irc_check_quit, (void*)from);
        }
        else if(j_strcmp(cmd, "PART") == 0) 
        {
            user u;
            chans channel;
            jpacket jp;
            xmlnode x;
            char *ptr;

            ptr = strchr(next, ' ');
            if(ptr != NULL)
                ptr[0] = '\0';

            channel = xhash_get(s->channels, make_lowercase(next));
            if(from == NULL || channel == NULL)
                continue;

            ptr = strchr(++from, '!');
            ptr[0] = '\0';

            u = xhash_get(channel->users, make_lowercase(from));
            if(u == NULL)
                continue;

            xhash_zap(channel->users, jid_full(u->id));
            xhash_zap(channel->users, u->user);

            jp = jpacket_new(xmlnode_new_tag("iq"));
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_BROWSE);
            xmlnode_put_attrib(jp->x, "to", jid_full(s->j));
            xmlnode_put_attrib(jp->x, "from", channel->id);
            xmlnode_put_attrib(jp->x, "type", "set");

            jpacket_reset(jp);
            x = xmlnode_insert_tag(jp->iq, "user");
            xmlnode_put_attrib(x, "jid", jid_full(u->id));
            xmlnode_put_attrib(x, "type", "remove");
            deliver(dpacket_new(jp->x), NULL);
        }
    }

    pool_free(p);
}

void irc_parse_phase_one(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    irctrans_data ircdata = (irctrans_data)s->data;
    char *line;
    pool p;

    log_debug(ZONE, "in phase one: m%X flag:%d, arg:%X, buffer: %s, bufsz:%d", m, flag, arg, buffer, bufsz);
    if(flag == MIO_ERROR)
        return;

    if(flag == MIO_CLOSED)
    {
        xmlnode_put_attrib(jp->x, "type", "error");
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "error"), "Socket Error to IRC Server", -1);
        xmlnode_put_attrib(xmlnode_get_tag(jp->x, "error"), "code", "502");
        deliver(dpacket_new(jp->x), NULL);
        xhash_zap(ircdata->sessions, spools(s->p, jid_full(s->j), "/", s->server, s->p));
        xhash_walk(s->channels, _irc_cleanup_channel, NULL);
        xhash_free(s->channels);
        if(s->buffer != NULL)
            free(s->buffer);
        pool_free(s->p);
        return;
    }

    irc_add_to_buffer(s, buffer, bufsz);

    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    {
        char *from = NULL, *cmd = NULL, *next = NULL;

        if(*line == '\0')
            continue;

        if(*line != ':')
        {
            cmd = line;
        }
        else
        {  
            from = line;
            cmd = strchr(line, ' ');
            if(cmd == NULL)
                continue;
            cmd[0] = '\0';
            cmd++;
        }

        next = strchr(cmd, ' ');
        if(next == NULL)
            continue;

        next[0] = '\0';
        next++;

        if(j_strcmp(cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", next, "\n", p), -1);
        }
        if(j_strcmp(cmd, "JOIN") == 0)
        {
            jpacket new;
            /* send out iq:result, and work on iq browse packet */
            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_CONFERENCE);
            jpacket_reset(jp);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "name"), s->nick, -1);
            s->nick_id = jid_new(s->p, jid_full(jp->from));
            jid_set(s->nick_id, shahash(s->nick), JID_RESOURCE);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "id"), jid_full(s->nick_id), -1);

            new = jpacket_new(xmlnode_new_tag("iq"));
            xmlnode_put_attrib(new->x, "type", "set");
            xmlnode_put_attrib(new->x, "from", jid_full(jp->from));
            xmlnode_put_attrib(new->x, "to", jid_full(s->j));
            xmlnode_put_attrib(xmlnode_insert_tag(new->x, "conference"), "xmlns", NS_BROWSE);
            jpacket_reset(new);
            xmlnode_put_attrib(new->iq, "type", "irc");
            xmlnode_put_attrib(new->iq, "jid", jid_full(jp->from));
            deliver(dpacket_new(jp->x), NULL);

            new->aux1 = s;
            jp = new;
            mio_reset(m, irc_parse_phase_one, (void*)new);
        }
        else if(j_strcmp(cmd, "353") == 0)
        {
            char *ptr;
            chans channel;
            /* :me 353 mynick = #channel :@me you them */
            
            ptr = strchr(next, '=');
            if(ptr == NULL)
                continue;
            next = ++ptr;
            ptr = ++next;
            next = strchr(ptr, ' ');
            if(next == NULL)
                continue;
            next[0] = '\0';

            channel = xhash_get(s->channels, make_lowercase(ptr));
            if(channel == NULL)
            {
                pool p_chan = pool_new();
                channel = pmalloco(p_chan, sizeof(_chans));
                channel->p = p_chan;
                channel->state = chan_JOINING;
                channel->name = irc_get_channel(p_chan, jid_full(jp->from));
                channel->id = pstrdup(p_chan, jid_full(jp->from));
                channel->users = xhash_new(101);
                channel->data = (void*)s;

                xhash_put(s->channels, channel->id, channel);
                xhash_put(s->channels, channel->name, channel);
            }

            ptr = ++next;
            ptr++;

            while(ptr != NULL)
            {
                pool p_user;
                user u;
                xmlnode x;
                next = strchr(ptr, ' ');
                if(next != NULL)
                {
                    next[0] = '\0';
                    next++;
                }

                p_user = pool_new();
                u = pmalloco(p_user, sizeof(_user));
                u->p = p_user;
                if(*ptr == '@')
                {
                    u->ops = 1;
                    ptr++;
                }
                else if(*ptr == '+')
                {
                    u->voice = 1;
                    ptr++;
                }
                u->user = make_lowercase(pstrdup(u->p, ptr));
                u->id   = jid_new(u->p, spools(p, jid_full(jp->from), "/", shahash(u->user), p));
                u->data = (void*)channel;

                ptr = next;
                if(ptr != NULL && *ptr == '\0')
                    ptr = NULL;

                xhash_put(channel->users, jid_full(u->id), u);
                xhash_put(channel->users, u->user, u);
                x = xmlnode_insert_tag(jp->iq, "user");
                xmlnode_put_attrib(x, "jid", jid_full(u->id));
                xmlnode_put_attrib(x, "name", u->user);
            }
        }
        else if(j_strcmp(cmd, "366") == 0) /* end of /NAMES list */
        {
            deliver(dpacket_new(jp->x), NULL);
            mio_reset(m, irc_parse_phase_two, (void*)s);
            /* kickstart phase two */
            mio_write(m, NULL, "PING :foo\n", -1);
        }
    }

    pool_free(p);
}


void irc_parse_phase_zero(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    irctrans_data ircdata = (irctrans_data)s->data;
    char *line;
    pool p;

    log_debug(ZONE, "in phase zero: m%X flag:%d, arg:%X, buffer: %s, bufsz:%d", m, flag, arg, buffer, bufsz);
    if(flag == MIO_ERROR)
    {
        xmlnode x;
        while((x = mio_cleanup(m)) != NULL)
            deliver_fail(dpacket_new(x), "Socket Error to IRC Server");
        return;
    }

    if(flag == MIO_NEW)
    {
        char *nick;
        s->m = m;

        nick = xmlnode_get_tag_data(jp->x, "query/nick");
        if(nick == NULL)
            nick = jp->from->user;

        xmlnode_hide(xmlnode_get_tag(jp->x, "query/nick"));

        mio_write(m, NULL, spools(s->p, "USER ", s->nick, " foo foo jabber_user@jabber.org\n", s->p), -1);
        mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n",  s->p), -1);
        return;
    }

    if(flag == MIO_CLOSED)
    {
        xhash_zap(ircdata->sessions, spools(s->p, jid_full(s->j), "/", s->server, s->p));
        xmlnode_put_attrib(jp->x, "type", "error");
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "error"), "Socket Error to IRC Server", -1);
        xmlnode_put_attrib(xmlnode_get_tag(jp->x, "error"), "code", "502");
        deliver(dpacket_new(jp->x), NULL);
        xhash_walk(s->channels, _irc_cleanup_channel, NULL);
        xhash_free(s->channels);
        if(s->buffer != NULL)
        {
            free(s->buffer);
            s->buffer = NULL;
        }
        pool_free(s->p);
        return;
    }

    irc_add_to_buffer(s, buffer, bufsz);

    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    {
        char *from = NULL, *cmd = NULL, *next = NULL;

        if(*line == '\0')
            continue;

        if(*line != ':')
        {
            cmd = line;
        }
        else
        {  
            from = line;
            cmd = strchr(line, ' ');
            if(cmd == NULL)
                continue;
            cmd[0] = '\0';
            cmd++;
        }

        next = strchr(cmd, ' ');
        if(next == NULL)
            continue;

        next[0] = '\0';
        next++;

        if(j_strcmp(cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", next, "\n", p), -1);
            continue;
        }
        else if(j_strcmp(cmd, "433") == 0) /* NICK IN USE */
        {
            char *nick = xmlnode_get_tag_data(jp->x, "query/nick");
            
            if(nick == NULL)
            {
                jutil_error(jp->x, TERROR_CONFLICT);
                deliver(dpacket_new(jp->x), NULL);
                mio_close(m);
                return;
            }
            else
            {
                mio_write(m, NULL, spools(p, "NICK ", nick, "\n", p), -1);
                return;
            }
            
        }
        else if(j_strcmp(cmd, "001") == 0) /* good nick change, we can join the channel now */
        {
            char *ptr = strchr(next, ' ');
            if(ptr == NULL)
            {
                mio_close(m);
                pool_free(p);
                return;
            }
            ptr[0] = '\0';
            s->nick = pstrdup(s->p, next);
            mio_reset(m, irc_parse_phase_one, arg);
            mio_write(m, NULL, spools(p, "JOIN ", irc_get_channel(p, jid_full(jp->to)), " ", xmlnode_get_tag_data(jp->x, "query/password"), "\n", p), -1);
            return;
        }
        else if(j_strcmp(cmd, "NICK") == 0)
        {
        }
    }

    pool_free(p);
}

void irc_check_iq_get(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    irctrans_data ircdata;
    char *line, *nick;
    pool p;

    if(s != NULL) ircdata = (irctrans_data)s->data;

    if(flag == MIO_ERROR) return;

    if(flag == MIO_CLOSED)
    {
        if(jp->aux1 != NULL)
        {
            xmlnode_put_attrib(jp->x, "type", "error");
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "error"), "Socket Error to IRC Server", -1);
            xmlnode_put_attrib(xmlnode_get_tag(jp->x, "error"), "code", "502");
            deliver(dpacket_new(jp->x), NULL);
            pool_free(s->p);
        }
        else
        {
            deliver(dpacket_new(jp->x), NULL);
        }
        return;
    }

    if(flag == MIO_NEW)
    {
        nick = pmalloco(m->p, 15);
        snprintf(nick, 15, "temp_%X", m);
        mio_write(m, NULL, spools(s->p, "USER ", nick, " foo foo jabber_user@jabber.org\n", s->p), -1);
        mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n", s->p), -1);
        return;
    }

    irc_add_to_buffer(s, buffer, bufsz);

    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    {
        char *from = NULL, *cmd = NULL, *next = NULL;

        if(*line == '\0')
            continue;

        if(*line != ':')
        {
            cmd = line;
        }
        else
        {  
            from = line;
            cmd = strchr(line, ' ');
            if(cmd == NULL)
                continue;
            cmd[0] = '\0';
            cmd++;
        }

        next = strchr(cmd, ' ');
        if(next == NULL)
            continue;

        next[0] = '\0';
        next++;

        if(j_strcmp(cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", next, "\n", p), -1);
        }
        else if(j_strcmp(cmd, "433") == 0) /* NICK IN USE */
        {
            mio_close(m);
            pool_free(p);
            return;
        }
        else if(j_strcmp(cmd, "001") == 0)
        {
            mio_write(m, NULL, spools(p, "MODE ", irc_get_channel(p, jid_full(jp->from)), "\n", p), -1);
        }
        else if(j_strcmp(cmd, "403") == 0) /* MODE REPLY -- Channel not found */
        {
            pool_free(s->p);
            jp->aux1 = NULL;
            mio_close(m);
            pool_free(p);
            return;
        }
        else if(j_strcmp(cmd, "324") == 0) /* MODE REPLY */
        {
            char *ptr = strchr(next, ' ');
            next = strchr(ptr, ' ');
            next++;
            
            if(strchr(next, 'k') != NULL)
                xmlnode_insert_tag(jp->iq, "password");

            pool_free(s->p);
            jp->aux1 = NULL;
            mio_close(m);
            pool_free(p);
            return;
        }
            
    }

    pool_free(p);
}

void irc_check_browse_get(mio m, int flag, void *arg, char *buffer, int bufsz)
{
    jpacket jp = (jpacket)arg;
    irc_session s = (irc_session)jp->aux1;
    irctrans_data ircdata;
    char *line, *nick;
    pool p;

    if(s != NULL) ircdata = (irctrans_data)s->data;

    if(flag == MIO_ERROR) return;

    if(flag == MIO_CLOSED)
    {
        /* return the packet, w/ or w/o the browse results */
        deliver(dpacket_new(jp->x), NULL);
        if(s->buffer != NULL) 
        {
            free(s->buffer);
            s->buffer = NULL;
        }
        pool_free(s->p);
        return;
    }

    if(flag == MIO_NEW)
    {
        nick = pmalloco(m->p, 15);
        snprintf(nick, 15, "temp_%X", m);
        mio_write(m, NULL, spools(s->p, "USER ", nick, " foo foo jabber_user@jabber.org\n", s->p), -1);
        mio_write(m, NULL, spools(s->p, "NICK ", nick, "\n", s->p), -1);
        return;
    }

    irc_add_to_buffer(s, buffer, bufsz);

    p = pool_new();

    while((line = irc_get_next_line(s, p)) != NULL)
    {
        char *from = NULL, *cmd = NULL, *next = NULL;

        if(*line == '\0')
            continue;

        if(*line != ':')
        {
            cmd = line;
        }
        else
        {  
            from = line;
            cmd = strchr(line, ' ');
            if(cmd == NULL)
                continue;
            cmd[0] = '\0';
            cmd++;
        }

        next = strchr(cmd, ' ');
        if(next == NULL)
            continue;

        next[0] = '\0';
        next++;

        if(j_strcmp(cmd, "PING") == 0)
        {
            mio_write(m, NULL, spools(p, "PONG ", next, "\n", p), -1);
        }
        else if(j_strcmp(cmd, "433") == 0) /* NICK IN USE */
        {
            mio_close(m);
            pool_free(p);
            return;
        }
        else if(j_strcmp(cmd, "001") == 0)
        {
            if(jp->from->user == NULL)
                mio_write(m, NULL, "LIST\n", -1);
            else
                mio_write(m, NULL, spools(p, "LIST :", jp->from->user, "\n", p), -1);
        }
        else if(j_strcmp(cmd, "322") == 0) /* channel listing */
        {
            char *channel;
            char *users;
            char *ptr;
            xmlnode x;

            channel = strchr(next, ' ');
            next = strchr(++channel, ' ');
            next[0] = '\0';
            users = ++next;
            next = strchr(users, ' ');
            next[0] = '\0';
            next += 2;
            ptr = strpbrk(next, "&<>\"\':");
            if(ptr != NULL) ptr[0] = '\0';
            next = str_to_UTF8(p, next);
            irc_strip_mirc_codes(next);

            if(jp->from->user == NULL)
                x = xmlnode_insert_tag(jp->iq, "conference");
            else
            {
                xmlnode_put_attrib((x = xmlnode_insert_tag(jp->x, "conference")), "xmlns", NS_BROWSE);
                jpacket_reset(jp);
                jp->aux1 = s;
            }
            xmlnode_put_attrib(x, "type", "irc");
            xmlnode_put_attrib(x, "jid", spools(p, channel, "@", ircdata->i->id, p));
            xmlnode_put_attrib(x, "name", spools(p, channel, " (", users, ") :", next, p));
        }
        else if(j_strcmp(cmd, "323") == 0) /* MODE REPLY */
        {
            if(jp->from->user == NULL) /* server request */
            {
                mio_close(m);
                pool_free(p);
                return;
            }

            /* room request */
            mio_write(m, NULL, spools(p, "WHO :", jp->from->user, "\n", p), -1);
        }
        else if(j_strcmp(cmd, "352") == 0)
        {
            char *ptr;
            xmlnode x = xmlnode_insert_tag(jp->iq, "user");

            /* 315 my_nick #channel user host server nick bla :name */
            ptr = strchr(next, '#'); /* channel */
            next = strchr(ptr, ' '); /* user */
            ptr = strchr(++next, ' '); /* host */
            next = strchr(++ptr, ' '); /* server */
            ptr = strchr(++next, ' '); /*nick*/
            next = strchr(++ptr, ' '); /* bla */
            if(next != NULL) 
                next[0] = '\0';
            else
                continue;
            
            xmlnode_put_attrib(x, "name", ptr);
            xmlnode_put_attrib(x, "jid", spools(p, jid_full(jp->from), "/", shahash(ptr), p));
        }
        else if(j_strcmp(cmd, "315") == 0)
        { /* send it on it's emrry way */
            mio_close(m);
            pool_free(p);
            return;
        }
    }

    pool_free(p);
}
